from django.conf import settings
from django.core.exceptions import ValidationError
from django.db import models
from django.utils import timezone
from django.utils.html import escape, linebreaks, urlize

import markdown2
from docutils.core import publish_parts
from markupfield.fields import MarkupField


def render_md(markup):
    return markdown2.markdown(markup, extras=[
        'code-friendly',
        'cuddled-lists',
        'fenced-code-blocks',
        'footnotes',
        'tables',
    ])


def render_rest(markup):
    parts = publish_parts(source=markup, writer_name="html4css1")
    return parts["fragment"]


MARKUP_FIELD_TYPES = [
    ('Markdown', render_md),
    ('Markdown Basic', markdown2.markdown),
    ('Plain Text', lambda markup: urlize(linebreaks(escape(markup)))),
    ('reStructuredText', render_rest),
]

STATUS_CHOICES = [
    ('Complete', 'Complete'),
    ('In Progress', 'In Progress'),
    ('Failed', 'Failed'),
]


class Hardware(models.Model):

    name = models.CharField(
        max_length=255,
        help_text="Enter the type of hardware.",
        unique=True,
    )

    class Meta:
        ordering = ['name']
        verbose_name_plural = 'hardware'

    def __str__(self):
        return self.name


class DocumentationRecord(models.Model):

    title = models.CharField(
        max_length=255,
        help_text="Enter a brief, descriptive title for this documentation.",
        unique=True,
    )

    maintenance_type = models.ForeignKey(
        'MaintenanceType',
        help_text='Select/Create a maintenance type.',
        on_delete=models.PROTECT,
    )

    documentation = MarkupField(
        blank=True,
        default_markup_type='Markdown',
        help_text='Document how to perform a task.',
        markup_choices=MARKUP_FIELD_TYPES,
        null=True,
    )

    created_at = models.DateTimeField(auto_now_add=True)
    updated_at = models.DateTimeField(auto_now=True)

    class Meta:
        ordering = ['title']

    def __str__(self):
        return self.title



class MaintenanceRecord(models.Model):
    system = models.ForeignKey(
        'System',
        help_text='Select/Create a system.',
        on_delete=models.PROTECT,
    )

    sys_admin = models.ForeignKey(
        'SysAdmin',
        help_text='Select a system administrator.',
        on_delete=models.PROTECT,
    )

    maintenance_type = models.ForeignKey(
        'MaintenanceType',
        help_text='Select/Create a maintenance type.',
        on_delete=models.PROTECT,
    )

    hardware = models.ManyToManyField(
        'Hardware',
        blank=True,
        help_text='Select the hardware involved in the system maintenance.',
    )

    software = models.ManyToManyField(
        'Software',
        blank=True,
        help_text='Select the software involved in the system maintenance.',
    )

    description = MarkupField(
        blank=True,
        default_markup_type='Markdown',
        help_text='Enter a description of the system maintenance performed.',
        markup_choices=MARKUP_FIELD_TYPES,
        null=True,
    )

    procedure = MarkupField(
        blank=True,
        default_markup_type='Markdown',
        help_text='Enter details of how the system maintenance was performed.',
        markup_choices=MARKUP_FIELD_TYPES,
        null=True,
    )

    problems = MarkupField(
        blank=True,
        default_markup_type='Markdown',
        help_text='Describe problems that arose during system maintenance.',
        markup_choices=MARKUP_FIELD_TYPES,
        null=True,
    )

    referenced_records = models.ManyToManyField(
        'self',
        related_name='referencing_records',
        symmetrical=False,
        through='MaintenanceRecordRelationship',
    )

    documentation_records = models.ManyToManyField(
        'DocumentationRecord',
        blank=True,
        help_text='Select documentation relevant to this system maintenance.<br>',
        related_name='maintenance_records',
    )

    status = models.CharField(
        choices = STATUS_CHOICES,
        default='In Progress',
        help_text='What is the current status of the system maintenance?',
        max_length=15,
    )

    datetime = models.DateTimeField(
        'maintenance date/time',
        default=timezone.now,
        help_text='Specify the date/time that the system maintenance was '
                  'performed.',
    )

    class Meta:
        ordering = ['-datetime']
        verbose_name = 'maintenance record'
        verbose_name_plural = 'maintenance records'

    def __str__(self):
        return '{} - {} ({})'.format(
            self.system, self.maintenance_type, self.datetime.date())


class MaintenanceRecordRelationship(models.Model):

    referencing_record = models.ForeignKey(
        'MaintenanceRecord',
        related_name='referencing_record',
        on_delete=models.CASCADE,
    )

    referenced_record = models.ForeignKey(
        'MaintenanceRecord',
        related_name='referenced_record',
        on_delete=models.CASCADE,
    )

    class Meta:
        ordering = ['referencing_record']
        unique_together = ('referencing_record', 'referenced_record')

    def clean(self):
        if self.referencing_record == self.referenced_record:
            raise ValidationError('A record cannot by related to itself.')

    def __str__(self):
        return '{} ➤ {}'.format(
            self.referencing_record, self.referenced_record)


class MaintenanceType(models.Model):

    maintenance_type = models.CharField(
        help_text="Enter a type of maintenance "
                  "(e.g., 'Software Installation').",
        max_length=255,
        unique=True,
    )

    description = models.TextField(
        blank=True,
        help_text='Enter a description of the maintenance type.',
    )

    class Meta:
        ordering = ['maintenance_type']

    def __str__(self):
        return self.maintenance_type


class Software(models.Model):

    name = models.CharField(
        max_length=255,
        help_text="Enter the software's name.",
        unique=True,
    )

    class Meta:
        ordering = ['name']
        verbose_name = 'software package'
        verbose_name_plural = 'software packages'

    def __str__(self):
        return self.name


class SysAdmin(models.Model):

    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        help_text='Select a user.',
        on_delete=models.PROTECT,
        unique=True,
    )

    class Meta:
        verbose_name = 'system administrator'
        verbose_name_plural = 'system administrators'

    def __str__(self):
        if self.user.first_name or self.user.last_name:
            return '{} ({})'.format(
                ' '.join([self.user.first_name, self.user.last_name]),
                self.user.username)
        else:
            return self.user.username


class System(models.Model):

    name = models.CharField(
        'system name',
        help_text='Enter a brief, unique identifier for the system.',
        max_length=255,
        unique=True,
    )

    description = models.TextField(
        blank=True,
        help_text='Enter a description of the system.',
    )

    class Meta:
        ordering = ['name']

    def __str__(self):
        return self.name



from django.contrib.auth.decorators import user_passes_test
from django.core.urlresolvers import reverse_lazy
from django.shortcuts import render
from django.utils.decorators import method_decorator
from django.views.generic import DetailView, ListView

from .app_settings import SYSTEM_MAINTENANCE_PAGINATE_BY
from .models import (DocumentationRecord, Hardware, MaintenanceRecord,
    MaintenanceType, Software, SysAdmin, System)


def sysadmin_check(user):
    """
    Check whether user is a sysadmin and has an active account.
    """

    return user.is_active and hasattr(user, 'sysadmin')


class SysAdminRequiredMixin(object):
    """
    Checks whether user is a sysadmin and has an active account.
    """

    @method_decorator(user_passes_test(
        sysadmin_check,
        login_url=reverse_lazy('system_maintenance:authentication')))
    def dispatch(self, *args, **kwargs):
        return super(SysAdminRequiredMixin, self).dispatch(*args, **kwargs)


@user_passes_test(
    sysadmin_check,
    login_url=reverse_lazy('system_maintenance:authentication'))
def raw_view(request, **kwargs):
    if kwargs['type_of_record'] == 'documentation':
        record = DocumentationRecord.objects.get(pk=kwargs['record_pk'])
    elif kwargs['type_of_record'] == 'maintenance':
        record = MaintenanceRecord.objects.get(pk=kwargs['record_pk'])
    else:
        pass

    field = getattr(record, kwargs['type_of_field'])

    context = {
        'type_of_field': kwargs['type_of_field'],
        'raw': field.raw,
        'record': record,
    }
    return render(
        request, 'system_maintenance/raw.html', context)


@user_passes_test(
    sysadmin_check,
    login_url=reverse_lazy('system_maintenance:authentication'))
def system_maintenance_home_view(request):
    context = {
        'documentation_record_count':
            DocumentationRecord.objects.all().count(),
        'hardware_count': Hardware.objects.all().count(),
        'maintenance_record_count': MaintenanceRecord.objects.all().count(),
        'maintenance_type_count': MaintenanceType.objects.all().count(),
        'software_count': Software.objects.all().count(),
        'sys_admin_count': SysAdmin.objects.all().count(),
        'system_count': System.objects.all().count(),
    }
    return render(
        request, 'system_maintenance/system_maintenance_home.html', context)


class DocumentationRecordDetailView(SysAdminRequiredMixin, DetailView):

    model = DocumentationRecord
    template_name = 'system_maintenance/documentation_record_detail.html'


class DocumentationRecordListView(SysAdminRequiredMixin, ListView):

    model = DocumentationRecord
    paginate_by = SYSTEM_MAINTENANCE_PAGINATE_BY
    template_name = 'system_maintenance/documentation_record_list.html'


class MaintenanceRecordDetailView(SysAdminRequiredMixin, DetailView):

    model = MaintenanceRecord
    template_name = 'system_maintenance/maintenance_record_detail.html'


class MaintenanceRecordListView(SysAdminRequiredMixin, ListView):

    model = MaintenanceRecord
    paginate_by = SYSTEM_MAINTENANCE_PAGINATE_BY
    template_name = 'system_maintenance/maintenance_record_list.html'



import os
import sys
from setuptools import setup

if sys.version_info < (3, 2):
    print("Sorry, django-system-maintenance currently requires Python 3.2+.")
    sys.exit(1)

# From: https://hynek.me/articles/sharing-your-labor-of-love-pypi-quick-and-dirty/
def read(*paths):
    """Build a file path from *paths* and return the contents."""
    with open(os.path.join(*paths), 'r') as f:
        return f.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

install_requires = [
    "Django>=1.7,<2.0",    # Confirmed good through 1.11.16
    "django-markupfield>=1.3.5",    # Confirmed good through 1.5.0
    "docutils>=0.12",    # Confirmed good through 0.14
    "markdown2>=2.3.0",    # Confirmed good through 2.3.6
    "Pygments>=2.0.2",    # Confirmed good through 2.2.0
]

setup(
    name='django-system-maintenance',
    version='0.3.2',
    packages=['system_maintenance'],
    include_package_data=True,
    license='BSD License',
    keywords='sysadmin system administration documentation maintenance',
    description='A Django app to document and track the administration and maintenance of computer systems',
    long_description=(read('README.rst') + '\n\n' +
                      read('CHANGELOG.rst')),
    url='https://github.com/mfcovington/django-system-maintenance',
    author='Michael F. Covington',
    author_email='mfcovington@gmail.com',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Web Environment',
        'Framework :: Django',
        'Framework :: Django :: 1.7',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3 :: Only',
        'Topic :: Documentation',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: System :: Systems Administration',
    ],
    install_requires=install_requires,
)
